module LoggingElf
  class TracingLogger
    extend Forwardable
    def_delegators :@logger, :<<, :<=>, :_dump_configuration, :_meta_eval,
      :add, :add_appenders, :additive, :additive=, :appenders, :appenders=,
      :clear_appenders, :inspect, :level, :level=, :name, :parent, :_setup,
      :remove_appenders, :trace, :trace=, :write, :formatter, :root, :config

    Logging.init

    def_delegators :@logger,
      *Logging::LEVELS.map { |name, _| "#{name}?".to_sym }

    Logging::LEVELS.each do |name, _|
      define_method name.to_sym do |data = nil|
        begin
          # data = yield if block_given?
          return if data.nil?
          data = append_trace_info(data)
          @logger.send(name.to_sym, data)
        rescue => err
          p err
          pp err.backtrace
        end
      end
    end

    def initialize(logger, &trace_hash)
      @logger = Logging::Logger.new logger if logger.is_a? String
      self.trace_hash = if block_given?
                          trace_hash
                        elsif LoggingElf.config.trace_hash
                          LoggingElf.config.trace_hash
                        else
                          fail "TracingLogger cannot be created with no" \
                            " mechanism for appending trace data"
                        end
    end

    attr_writer :trace_hash
    def trace_hash
      hash = if @trace_hash.respond_to?(:call)
               @trace_hash.call
             else
               @trace_hash
             end

      hash || {}
    end

    def exception_data(data)
      {
        error_object: data,
        backtrace: data.backtrace,
        short_message: "Exception: #{data.message}",
        message: data.message,
        is_exception: true
      }
    end

    def default_data(data)
      {
        short_message: "Unknown thing to log",
        message: data.inspect
      }
    end

    def append_trace_info(data)
      data = case data
             when Hash then data
             when Exception then exception_data(data)
             when String then { message: data }
             when nil then return
             else
               default_data(data)
             end
      data.merge(trace_hash)
    end
  end
end
